from celery import shared_task


@shared_task(default_retry_delay=5)
def index(content_type_id, pk, refresh=False):
    from django.contrib.contenttypes.models import ContentType
    content_type = ContentType.objects.get_for_id(content_type_id)
    obj = content_type.model_class().objects.get(id=pk)
    obj.index(refresh=refresh)



from rest_framework import serializers

from bulbs.utils.serializers import JSONField

from .models import SpecialCoverage


class SpecialCoverageSerializer(serializers.ModelSerializer):

    query = JSONField(required=False, default={})
    videos = JSONField(required=False, default=[])
    config = JSONField(required=False, default={})

    class Meta:
        model = SpecialCoverage



from django.conf import settings

from .models import Contribution, ContributorRole, FEATURETYPE


def get_forced_payment_contributions(start_date, end_date, qs=None):
    if not qs:
        qs = Contribution.objects.filter(force_payment=True)
    include = qs.filter(
        payment_date__range=(start_date, end_date)
    ) | qs.filter(
        payment_date__isnull=True
    )
    exclude = qs.exclude(
        payment_date__isnull=True
    ).exclude(
        payment_date__range=(start_date, end_date),
    )
    return include, exclude


def update_content_contributions(content, author):
    if not Contribution.objects.filter(content=content, contributor=author).exists():
        role_name = getattr(settings, 'DEFAULT_CONTRIBUTOR_ROLE', 'default')
        role, created = ContributorRole.objects.get_or_create(name=role_name)
        if created:
            role.payment_type = FEATURETYPE
            role.save()
        Contribution.objects.create(content=content, contributor=author, role=role)



from django.conf import settings
from django.db import models

from bulbs.content.models import Content, FeatureType


FLAT_RATE = 0
FEATURETYPE = 1
HOURLY = 2
MANUAL = 3
OVERRIDE = 4

ROLE_PAYMENT_TYPES = (
    (FLAT_RATE, 'Flat Rate'),
    (FEATURETYPE, 'FeatureType'),
    (HOURLY, 'Hourly'),
    (MANUAL, 'Manual')
)

RATE_PAYMENT_TYPES = ROLE_PAYMENT_TYPES + ((OVERRIDE, 'Override'),)


def calculate_hourly_pay(rate, minutes_worked):
    return ((float(rate) / 60) * minutes_worked)


class LineItem(models.Model):
    contributor = models.ForeignKey(settings.AUTH_USER_MODEL)
    amount = models.IntegerField(default=0)
    note = models.TextField()
    payment_date = models.DateTimeField(null=True, blank=True)


class ContributorRole(models.Model):
    name = models.CharField(max_length=255)
    description = models.TextField(null=True, blank=True)
    payment_type = models.IntegerField(choices=ROLE_PAYMENT_TYPES, default=MANUAL)

    def save(self, *args, **kwargs):
        super(ContributorRole, self).save(*args, **kwargs)
        if self.payment_type == 1:
            self.create_feature_type_rates()

    def create_feature_type_rates(self):
        for feature_type in FeatureType.objects.all():
            rate = FeatureTypeRate.objects.filter(role=self, feature_type=feature_type)
            if not rate.exists():
                FeatureTypeRate.objects.create(role=self, feature_type=feature_type, rate=0)

    def get_rate(self):
        if self.payment_type == FLAT_RATE:
            qs = self.flat_rates.all()
            if qs.exists():
                return qs.first()
        if self.payment_type == HOURLY:
            qs = self.hourly_rates.all()
            if qs.exists():
                return qs.first()
        return None


class Contribution(models.Model):
    role = models.ForeignKey(ContributorRole)
    contributor = models.ForeignKey(settings.AUTH_USER_MODEL, related_name="contributions")
    content = models.ForeignKey(Content, related_name="contributions")
    notes = models.TextField(null=True, blank=True)
    minutes_worked = models.IntegerField(null=True)
    force_payment = models.BooleanField(default=False)
    payment_date = models.DateTimeField(null=True, blank=True)

    @property
    def get_pay(self):
        return self._get_pay()

    @property
    def get_override(self):
        return self._get_override()

    def get_rate(self):
        payment_type = self.role.payment_type
        if payment_type == MANUAL:
            return self.manual_rates.all().first()

        if payment_type == FLAT_RATE:
            return self.role.flat_rates.filter().first()

        if self.content.feature_type and payment_type == FEATURETYPE:
            return self.content.feature_type.feature_type_rates.filter(
                role=self.role
            ).first()

        if payment_type == HOURLY:
            return self.role.hourly_rates.filter().first()

    def _get_override(self):
        # Get contribution specific overrides first.
        if self.role.payment_type in [0, 1] and self.override_contribution.exists():
            return self.override_contribution.first().rate

        override_profile = self.contributor.overrides.filter(role=self.role).first()
        if not override_profile:
            return None

        elif self.role.payment_type == 0 and override_profile.override_flatrate.exists():
            return override_profile.override_flatrate.first().rate

        elif (self.role.payment_type == 1):
            override = override_profile.override_feature_type.filter(
                feature_type=self.content.feature_type
            ).first()
            if override:
                return override.rate

        elif self.role.payment_type == 2 and override_profile.override_hourly.exists():
            minutes_worked = float(getattr(self, 'minutes_worked', 0))
            return calculate_hourly_pay(
                override_profile.override_hourly.first().rate, minutes_worked
            )

    def _get_pay(self):
        override = self.get_override
        if override:
            return override
        rate = self.get_rate()
        if isinstance(rate, HourlyRate):
            minutes_worked = float(getattr(self, 'minutes_worked', 0))
            return calculate_hourly_pay(rate.rate, minutes_worked)
        if rate:
            return rate.rate
        return None


class OverrideProfile(models.Model):
    contributor = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='overrides')
    role = models.ForeignKey(ContributorRole, related_name='overrides')

    class Meta:
        unique_together = (('contributor', 'role'),)


class BaseOverride(models.Model):
    rate = models.IntegerField(default=0)
    updated_on = models.DateTimeField(auto_now_add=True)

    class Meta:
        ordering = ('-updated_on',)


class FeatureTypeOverride(BaseOverride):
    profile = models.ForeignKey(OverrideProfile, related_name='override_feature_type')
    feature_type = models.ForeignKey(FeatureType)


class FlatRateOverride(BaseOverride):
    profile = models.ForeignKey(OverrideProfile, related_name='override_flatrate')


class HourlyOverride(BaseOverride):
    profile = models.ForeignKey(OverrideProfile, related_name='override_hourly')


class ContributionOverride(BaseOverride):
    contribution = models.ForeignKey(Contribution, related_name='override_contribution')


class Rate(models.Model):
    name = models.IntegerField(choices=RATE_PAYMENT_TYPES, null=True)
    updated_on = models.DateTimeField(auto_now=True)
    rate = models.IntegerField()

    class Meta:
        ordering = ('-updated_on',)


class FlatRate(Rate):
    role = models.ForeignKey(ContributorRole, related_name="flat_rates")


class HourlyRate(Rate):
    role = models.ForeignKey(ContributorRole, related_name="hourly_rates")


class ManualRate(Rate):
    contribution = models.ForeignKey(Contribution, related_name="manual_rates")


class FeatureTypeRate(Rate):
    role = models.ForeignKey(ContributorRole, null=True, related_name="feature_type_rates")
    feature_type = models.ForeignKey(FeatureType, related_name="feature_type_rates")

    class Meta:
        unique_together = (("role", "feature_type"))


class FreelanceProfile(models.Model):
    contributor = models.OneToOneField(settings.AUTH_USER_MODEL, unique=True)
    is_freelance = models.BooleanField(default=True)
    is_manager = models.BooleanField(default=False)
    payment_date = models.DateTimeField(null=True, blank=True)

    def get_pay(self, start=None, end=None):
        qs = self.contributor.contributions.all()
        if start:
            qs = qs.filter(payment_date__gte=start)
        if end:
            qs = qs.filter(payment_date__lte=end)

        pay = 0
        for contribution in qs.all():
            contribution_pay = contribution.get_pay
            if contribution_pay:
                pay += contribution_pay
        return pay



from celery.task import task

from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.db import models
from django.utils import timezone

from json_field import JSONField

from bulbs.content.models import Content
from .operations import *  # noqa


@task
def update_pzone(**kwargs):
    """Update pzone data in the DB"""

    pzone = PZone.objects.get(**kwargs)

    # get the data and loop through operate_on, applying them if necessary
    when = timezone.now()
    data = pzone.data
    for operation in pzone.operations.filter(when__lte=when, applied=False):
        data = operation.apply(data)
        operation.applied = True
        operation.save()
    pzone.data = data

    # create a history entry
    pzone.history.create(data=pzone.data)

    # save modified pzone, making transactions permanent
    pzone.save()


class PZoneManager(models.Manager):

    def operate_on(self, when=None, apply=False, **kwargs):
        """Do something with operate_on. If apply is True, all transactions will
        be applied and saved via celery task."""

        # get pzone based on id
        pzone = self.get(**kwargs)

        # cache the current time
        now = timezone.now()

        # ensure we have some value for when
        if when is None:
            when = now

        if when < now:
            histories = pzone.history.filter(date__lte=when)
            if histories.exists():
                # we have some history, use its data
                pzone.data = histories[0].data

        else:
            # only apply operations if cache is expired or empty, or we're looking at the future
            data = pzone.data

            # Get the cached time of the next expiration
            next_operation_time = cache.get('pzone-operation-expiry-' + pzone.name)
            if next_operation_time is None or next_operation_time < when:

                # start applying operations
                pending_operations = pzone.operations.filter(when__lte=when, applied=False)
                for operation in pending_operations:
                    data = operation.apply(data)

                # reassign data
                pzone.data = data

                if apply and pending_operations.exists():
                    # there are operations to apply, do celery task
                    update_pzone.delay(**kwargs)

        # return pzone, modified if apply was True
        return pzone

    def preview(self, when=timezone.now(), **kwargs):
        """Preview transactions, but don't actually save changes to list."""

        return self.operate_on(when=when, apply=False, **kwargs)

    def applied(self, **kwargs):
        """Apply transactions via a background task, return preview to user."""

        return self.operate_on(apply=True, **kwargs)


class PZone(models.Model):
    name = models.SlugField(unique=True)
    zone_length = models.IntegerField(default=10)
    data = JSONField(default=[], blank=True)

    objects = PZoneManager()

    def __len__(self):
        return min(self.zone_length, len(self.data))

    def __iter__(self):
        content_ids = [item["id"] for item in self.data[:self.__len__()]]
        bulk = Content.objects.in_bulk(content_ids)
        for pk in content_ids:
            yield bulk.get(pk)

    def __getitem__(self, index):
        items = self.data[:self.__len__()].__getitem__(index)
        if isinstance(items, dict):
            return Content.objects.get(id=self.data[index]["id"])
        if isinstance(items, list):
            content = []
            content_ids = [item["id"] for item in items]
            bulk = Content.objects.in_bulk(content_ids)
            for pk in content_ids:
                content.append(bulk.get(pk))
            return content
        raise IndexError("Index out of range")

    def __setitem__(self, index, value):
        if index > self.__len__():
            raise IndexError("Index out of range")
        if isinstance(value, Content):
            self.data[index]["id"] = value.pk
        elif isinstance(value, int):
            self.data[index]["id"] = value
        else:
            raise ValueError("PZone items must be Content or int")

    def __delitem__(self, index):
        if index > self.__len__():
            raise IndexError("Index out of range")
        del self.data[index]

    def __contains__(self, value):
        if isinstance(value, Content):
            value = value.pk
        if isinstance(value, int):
            for item in self.data:
                if value == item["id"]:
                    return True
        return False

    def __unicode__(self):
        return "{}[{}]".format(self.name, self.__len__())

    def clean(self, *args, **kwargs):
        super(PZone, self).clean(*args, **kwargs)

        if not isinstance(self.data, list):
            raise ValidationError('PZone data must be formatted as a list')

        for instance in self.data:
            if not isinstance(instance, dict) or 'id' not in instance:
                raise ValidationError(
                    'PZone data objects must be formatted like the following "{id: int}"'
                )

    def save(self, *args, **kwargs):
        self.clean()
        super(PZone, self).save(*args, **kwargs)


    class Meta:
        ordering = ["name"]


class PZoneHistory(models.Model):
    pzone = models.ForeignKey(PZone, related_name="history")
    data = JSONField(default=[])
    date = models.DateTimeField(auto_now_add=True)

    class Meta:
        # we want the most recently created to come out first
        ordering = ["-date"]
